import numpy as np
from ezauv.simulation.animator import SimulationAnimator
from ezauv.simulation.fake_sensors import FakeDepthSensor, FakeIMU
from scipy.spatial.transform import Rotation as R
import quaternion

# kinda sucks, 2d
class Simulation:

    def __init__(self, motor_locations, motor_directions, inertia, bounds, deadzone):
        self.location = np.array([0., 0.])
        self.velocity = np.array([0., 0.])
        self.acceleration = np.array([0., 0.])
        self.rotation = 0 # degrees
        # start by facing positive x axis
        self.rotational_velocity = 0
        self.rotational_acceleration = 0

        self.motor_locations = motor_locations # relative to center
        self.motor_directions = motor_directions
        self.motor_speeds = np.array([np.array([0., 0., 0.]) for _ in motor_directions])
        # print(self.motor_speeds)
        self.moment_of_inertia = inertia

        self.timestep = 0.01 # in seconds
        self.prevtime = 0
        self.animator = SimulationAnimator(fps=1/self.timestep)

        self.drag = 0

        self.bounds = bounds
        self.deadzone = deadzone

        self.rng = np.random.default_rng(int(100))

        self.real_accel = np.array([0., 0.])
        self.prev_vel = np.array([0., 0.])
    
    def simulate(self, time):
        # set_text(str(self.rotation))
        
        for timepoint in np.arange(self.prevtime, time, self.timestep):
            self.real_accel = (self.velocity - self.prev_vel) / self.timestep
            self.prev_vel = self.acceleration

            self.velocity += (self.rng.random() - 0.5) * 0.01
            self.rotational_velocity += ((2 * (self.rng.random() - 0.5))) * 0.1
            self.location += self.velocity * self.timestep
            # print(self.acceleration)
            # print(self.timestep)
            self.velocity += self.acceleration * self.timestep
            self.rotation = (self.rotation + self.rotational_velocity * self.timestep) % 360

            self.rotational_velocity += self.rotational_acceleration * self.timestep
            rotation = R.from_euler('z', np.deg2rad(self.rotation))
            
            
            rotated_speeds = []
            rotated_locations = []
            torques = []
            for i in range(len(self.motor_speeds)):
                s = np.delete(rotation.apply(np.array([self.motor_speeds[i][0], self.motor_speeds[i][1], 0.0])), 2)
                l = np.delete(rotation.apply(np.array([self.motor_locations[i][0], self.motor_locations[i][1], 0.0])), 2)
                rotated_speeds.append(s)
                rotated_locations.append(l)
                torques.append(np.cross(l, s))
            self.acceleration = sum(rotated_speeds)
            torque = sum(torques)

            self.rotational_acceleration = torque / self.moment_of_inertia
            
            self.animator.append(
                self.location,
                self.rotation,
                self.velocity,
                [loc + self.location for loc in rotated_locations],
                rotated_speeds
            )
            self.rotational_velocity *= (1 - self.drag)
            self.velocity *= (1 - self.drag)

    def render(self):
        self.animator.render()

    def update_motor_speeds(self, speeds):
        self.motor_speeds = [d * s for d, s in zip(self.motor_directions, speeds)]
        

    def depth(self, dev):
        return FakeDepthSensor(dev)
    
    def imu(self, dev):
        return FakeIMU(dev, lambda: self.real_accel, lambda: quaternion.from_euler_angles([np.deg2rad(self.rotation), 0., 0.]))
    
    def set(self, index, speed):
            
            speed = max(min(speed, self.bounds[index][1]), self.bounds[index][0])
            if(self.deadzone[index][0] > speed > self.deadzone[index][1]):
                speed = 0
            # set_text(str(speed))
            self.motor_speeds[index] = self.motor_directions[index] * speed
            # print(self.motor_speeds)
    
    def set_motor(self, index):
        return lambda speed: self.set(index, speed)
    
    def apply_force(self, *, thrust, rotation):
        self.velocity += thrust
        self.rotational_velocity += rotation
        

        

# a = Simulation(
#         np.array([
#             np.array([-1., -1.]),
#             np.array([-1., 1.]),
#             np.array([1., 1.]),
#             np.array([1., -1.])
#         ]),
#         np.array([
#                 np.array([1., -1.]),
#                 np.array([1., 1.]),
#                 np.array([1., -1.]),
#                 np.array([1., 1.])
#         ]),
#         1/6
# )
# # a.update_motor_speeds(np.array([1, 1, 1, 1]))
# # a.simulate(2)
# # a.update_motor_speeds(np.array([1, -1, -1, 1]))
# # a.simulate(8)
# # a.update_motor_speeds(np.array([0, 0, 0, 0]))
# # a.simulate(8)
# # # a.update_motor_speeds(np.array([0, 0, 0, 0]))
# # # a.simulate(6)
# # a.render()


# # # def emergency_kill():
#     # pass

# # def set_motor(id, speed):
# #     set_pin(id, speed)



from ezauv.mission.mission import Task
from ezauv.hardware.sensor_interface import SensorInterface

import numpy as np
import time

class AccelerateVector(Task):

    def __init__(self, vec, length):
        super().__init__()
        self.vec = vec
        self.length = length
        self.start = -1

    @property
    def name(self) -> str:
        return "Accelerate at vector task"
    
    @property
    def finished(self) -> bool:
        if(self.start == -1):
            self.start = time.time()
        return (time.time() - self.start) >= self.length

    def update(self, sensors: SensorInterface) -> np.ndarray:
        if(self.start == -1):
            self.start = time.time()
        return self.vec
        


import os
import time

from enum import StrEnum

class LogLevel(StrEnum):
    INFO = "    Info   "
    WARNING = "  Warning  "
    ERROR = "!!!ERROR!!!"
    DEBUG = "   Debug   "

class Logger:

    def __init__(self, console: bool, file: bool):
        
        self.console: bool = console
        self.file: bool = file

        self.dead = False

        if(self.file):
            os.makedirs("logs", exist_ok=True)
            self.location = f"logs/{time.asctime().replace(' ', '-')}"
            self.stream = open(self.location, "w", encoding="utf-8")

        self.log("   date   h:m:s:microsecond   source        level       message\n", info=False)

    def log(self, message: str, level: LogLevel = LogLevel.INFO, source: str = " GENERAL ", info: bool = True): 
        if(self.dead):
            self.stream = open(self.location, "w", encoding="utf-8")
            self.dead = False


        if(info):
            timestamp = time.time()
            seconds_timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp))
            microseconds = int((timestamp * 1000000) % 1000000)

            formatted_timestamp = f"{seconds_timestamp}:{microseconds:06d}"

            message = formatted_timestamp + " : " + source + " : " + level + " : " + message

        if(level == LogLevel.ERROR):
            message = "\n!!!ERROR!!!\n" + message + "\n!!!ERROR!!!\n"


        if(self.console):
            print(message)

        if(self.file):
            self.stream.write(message + "\n")

    def end(self):
        if(self.file):
            self.stream.close()
            self.dead = True
            

    def create_sourced_logger(self, source):
        if(self.dead):
            self.stream = open(self.location, "w", encoding="utf-8")
            self.dead = False

        if len(source) <= 9:
            extra = 9 - len(source)
            back = extra // 2

            source = (" " * back) + source + (" " * (extra - back))

        return lambda message, level=LogLevel.INFO: self.log(message, level, source)



from ezauv.mission.mission import Task
from ezauv.hardware.sensor_interface import SensorInterface

import numpy as np
import time

class RunFunction(Task):

    def __init__(self, func):
        super().__init__()
        self.func = func
        self.run = False

    @property
    def name(self) -> str:
        return "Run function task"
    
    @property
    def finished(self) -> bool:
        return self.run

    def update(self, sensors: SensorInterface) -> np.ndarray:
        self.func()
        self.run = True
        return np.array([0., 0., 0., 0., 0., 0.])
        


from typing import List, Callable, Optional
import numpy as np
from gurobipy import GRB, Model, quicksum
import quaternion
import time

from ezauv.utils.logger import LogLevel


class DeadzoneOptimizer:
    #TODO actually document any of this
    # ok just for future reference b4 documentation bounds are like (-1, 1) and deadzones are like (0.1, 0.1). 
    # deadzones do not go negative! might change in future. probably will use ranges from motors?
    # need 1 per motor, so both are lists
    def __init__(self, M, bounds, deadzones):
        self.M = M
        self.bounds = bounds
        self.deadzones = deadzones
        self.m, self.n = M.shape

        self.model = Model("MIQP_deadzone") # miqp = mixed integer quadratic programming
        # quadratic because it minimized the sum of squares of elements of the matrix
        # integer because it uses boolean variables to determine what side of the deadzone the continous variables are on
        # mixed because it also has continuous
        # programming meaning optimization, because it minimizes the sum of squares

        self.eps = self.model.addVars(self.m, lb=-float('inf'), vtype=GRB.CONTINUOUS, name="eps")

        self.u = {}
        for i in range(self.n):
            self.u[i] = self.model.addVar(lb=bounds[i][0], ub=bounds[i][1], vtype=GRB.CONTINUOUS, name=f"u_{i}")

        self.z = self.model.addVars(self.n, vtype=GRB.BINARY, name="z")
        self.s = self.model.addVars(self.n, vtype=GRB.BINARY, name="s")
        

        self.M0 = max(abs(b) for bound in bounds for b in bound)

        for i in range(self.n):
            self.model.addConstr(self.u[i] >= -self.z[i] * bounds[i][1], name=f"u_lower_bound_{i}")
            self.model.addConstr(self.u[i] <= self.z[i] * bounds[i][1], name=f"u_upper_bound_{i}")
            # either bounded in (-b, b) or (0, 0)

        
        for i in range(self.n):
            self.model.addGenConstrIndicator(self.z[i], 1,
                self.u[i] - deadzones[i][1] * self.s[i] + self.M0 * (1 - self.s[i]),
                GRB.GREATER_EQUAL, 0, name=f"deadzone_lower_{i}")

            self.model.addGenConstrIndicator(self.z[i], 1,
                self.u[i] - self.M0 * self.s[i] + deadzones[i][0] * (1 - self.s[i]),
                GRB.LESS_EQUAL, 0, name=f"deadzone_upper_{i}")


        self.constrs = []
        for j in range(self.m):
            expr = quicksum(self.M[j, i] * (self.u[i]) for i in range(self.n)) + self.eps[j]
            # expr = self.eps[j] - 1
            # expr = quicksum(M[j, i] * self.u[i] for i in range(self.n))
            # expr = 1 - self.eps[j]
            self.constrs.append(self.model.addConstr(expr == 0, name=f"eq_row_{j}"))
            # matrix multiplication must be true

        self.model.Params.OutputFlag = 0
        self.model.update()

        

    def optimize(self, V, M):
        for j in range(self.m):
            self.constrs[j].setAttr(GRB.Attr.RHS, V[j])
            for i in range(self.n):
                self.model.chgCoeff(self.constrs[j], self.u[i], M[j, i])

        self.model.setObjective(quicksum(self.eps[j] * self.eps[j] for j in range(self.m)), GRB.MINIMIZE)
        self.model.optimize()

        if self.model.status != GRB.OPTIMAL:
            return False, None


        eps_opt = [self.eps[j].X for j in range(self.m)]
        # print(eps_opt)
        # if(not np.all(np.isclose(eps_opt, 0.0))):
            # set_text("rotation: " + " ".join(str(t) for t in V))
            # print(V[5])
        
        for j in range(self.m):
            self.eps[j].LB = eps_opt[j]
            self.eps[j].UB = eps_opt[j]

        self.model.setObjective(quicksum(self.u[i] * self.u[i] for i in range(self.n)), GRB.MINIMIZE)
        self.model.optimize()

        for j in range(self.m):
            self.eps[j].LB = -float('inf')
            self.eps[j].UB = float('inf')
        if self.model.status == GRB.OPTIMAL:
            return True, np.array([self.u[i].X for i in range(self.n)])

        return False, None


test = []
class Motor:

    class Range:
        def __init__(self, bottom: float, top: float):
            self.max = top
            self.min = bottom

    def __init__(self, thrust_vector: np.ndarray, position: np.ndarray, set_motor: Callable, initialize: Callable, bounds: Range, deadzone: Range):
        self.thrust_vector: np.ndarray = thrust_vector
        self.position: np.ndarray = position
        self.set: Callable = set_motor

        self.initialize: Callable = initialize
        self.inertia_tensor: Optional[np.ndarray] = None
        self.torque_vector: Optional[np.ndarray] = None

        self.bounds: Motor.Range = bounds
        self.deadzone: Motor.Range = deadzone

    def set_inertia_tensor(self, inertia_tensor):
        self.inertia_tensor = inertia_tensor
        self.torque_vector = np.cross(self.position, self.thrust_vector)    


class MotorController:

    def __init__(self, *, inertia: np.ndarray, motors: List[Motor]):
        self.inertia: np.ndarray = inertia  # the inertia tensor of the entire body
        self.motors: np.ndarray = np.array(motors)   # the list of motors this sub owns
        self.log: Callable = lambda str, level=None: print(f"Motor logger is not set --- {str}")

        self.optimizer: Optional[DeadzoneOptimizer] = None

        for motor in motors:
            motor.set_inertia_tensor(self.inertia)

        self.motor_matrix = None
        self.mT = None
        self.reset_optimizer()

    def overview(self) -> None:
        self.log("---Motor controller overview---")
        self.log(f"Inertia tensor:\n{self.inertia}")
        self.log(f"{len(self.motors)} motors connected")

    def initialize(self) -> None:
        self.log("Initializing motors...")
 
        problems = 0
        for motor in self.motors:
            problems += motor.initialize()
        
        level = LogLevel.INFO if problems == 0 else LogLevel.WARNING

        self.log(f"Motors initalized with {problems} problem{'' if problems==1 else 's'}", level=level)
    
    def reset_optimizer(self):
        bounds = []
        deadzones = []

        for i, motor in enumerate(self.motors):
            # print(motor.thrust_vector)
            new_vector = np.array([np.concatenate([motor.thrust_vector, self.inertia @ motor.torque_vector], axis=None)]).T
            if(i == 0):
                self.motor_matrix = new_vector
            else:
                self.motor_matrix = np.hstack((self.motor_matrix, new_vector))

            bounds.append((motor.bounds.min, motor.bounds.max))
            deadzones.append((motor.deadzone.min, motor.deadzone.max))
        self.optimizer = DeadzoneOptimizer(self.motor_matrix, bounds, deadzones)
        self.mT = self.motor_matrix.T

    def rotate(self, rotation):
        rotated_vectors = []
        for vec in self.mT:
            split_vec = np.split(vec, [3])
            # a = []
            # a.extend([quaternion.rotate_vectors(rotation, split_vec)])
            # print(a)
            rotated_vectors.append([val for sublist in quaternion.rotate_vectors(rotation, split_vec) for val in sublist])

        return np.array(rotated_vectors).T

    def solve(self, wanted_vector, rotation):
        # print(self.motor_matrix)
        # print(self.rotate(rotation))
        # wanted_vector = 
        # print(self.optimizer.optimize(np.array([0., 0., 0., 0., 0., 2295.18311443701]), self.rotate(rotation)))
        # raise Exception()
        start = time.time()
        optimized = self.optimizer.optimize(wanted_vector, self.rotate(rotation))
        test.append(time.time() - start)
        # if(not optimized[0]):
        return optimized
    
    def set_motors(self, motor_speeds):
        # print(len(motor_speeds))
        # print(motor_speed)
        # print(self.motor_matrix)
        # print(motor_speeds)
        for i, motor in enumerate(self.motors):
            motor.set(motor_speeds[i])


# from inertia import *


# test = MotorController(inertia=InertiaBuilder(Cuboid(10, np.array([0, 0, 0]), 5, 5, 1)).moment_of_inertia(), 
#                        motors=[
#                             Motor(np.array([-1, 1, 0]), np.array([-1, -1, 0]), lambda num: print(num), lambda _: 0, Motor.Range(-0.2, 0.2), Motor.Range(0.11, 0.11)),
#                             Motor(np.array([1, 1, 0]), np.array([1, -1, 0]), lambda num: print(num), lambda _: 0, Motor.Range(-0.2, 0.2), Motor.Range(0.11, 0.11)),
#                             Motor(np.array([1, 1, 0]), np.array([-1, 1, 0]), lambda num: print(num), lambda _: 0, Motor.Range(-0.2, 0.2), Motor.Range(0.11, 0.11)),
#                             Motor(np.array([-1, 1, 0]), np.array([1, 1, 0]), lambda num: print(num), lambda _: 0, Motor.Range(-0.2, 0.2), Motor.Range(0.11, 0.11)),
#                             Motor(np.array([0, 0, 1]), np.array([0, 0.5, 0]), lambda num: print(num), lambda _: 0, Motor.Range(-0.2, 0.2), Motor.Range(0.11, 0.11)),
#                             Motor(np.array([0, 0, 1]), np.array([0, -0.5, 0]), lambda num: print(num), lambda _: 0, Motor.Range(-0.2, 0.2), Motor.Range(0.11, 0.11))
#                            ]
#                        )

# # target = np.array([0.9,0,0,0,0,0])
# # print(test.solve(target, quaternion.from_euler_angles(np.deg2rad(0), 0, 0)))
# # # test.set_motors(test.solve(target)[1])
# # # print(quaternion.rotate_vectors(np.quaternion(1,0,0,0), np.array([np.array([1,1,1]), np.array([1,1,1])])))
# # import time
# # avg = 0.
# # count = 0
# # for i in np.linspace(-1, 1, 1000):
# #     count += 1
# #     start = time.time()
# #     test.solve(np.array([i, 0, 0, 0, 0, 0]), quaternion.from_euler_angles(0, 0, 0))
# #     avg += time.time() - start

# # print(f"Average time taken: {(avg / count) * 1000} milliseconds")


import numpy as np
from abc import ABC, abstractmethod
import quaternion
import time


# The abstract interface for interacting with the hardware; this should be
# extended and registered into the auv

class ImuInterface(ABC):

    def __init__(self):
        super().__init__()
        self.log = lambda str, level=None: print(str)

    @abstractmethod
    def get_accelerations(self) -> np.ndarray:
        pass

    @abstractmethod
    def get_rotation(self) -> np.quaternion:
        pass

    @abstractmethod
    def initialize(self) -> None:
        pass

    @abstractmethod
    def overview(self) -> None:
        pass

class DepthInterface(ABC):

    def __init__(self):
        super().__init__()
        self.log = lambda str: print(str)
    
    @abstractmethod
    def get_depth(self) -> float:
        pass

    @abstractmethod
    def initialize(self) -> None:
        pass

    @abstractmethod
    def overview(self) -> str:
        pass

class SensorInterface:

    def __init__(self, *, imu: ImuInterface, depth: DepthInterface):
        self.imu: ImuInterface = imu
        self.depth: DepthInterface = depth
        self.log = lambda str: print(str)

    def initialize(self) -> None:
        self.imu.initialize()
        self.depth.initialize()

    def overview(self) -> None:
        self.imu.overview()
        self.depth.overview()

    def killed(self, average_time=1, max_dev=0.01, timestep=0.001) -> bool:
        avg = 0
        total = int(np.ceil(average_time / timestep))

        for i in range(total):
            avg += self.imu.get_accelerations()
            time.sleep(timestep)

        avg /= total
        return np.all(abs(avg) <= max_dev)



from abc import ABC, abstractmethod
import numpy as np

class InertiaGeometry(ABC):

    def __init__(self, mass: int, center: np.ndarray):
        self.mass = mass
        self.center = center

    @abstractmethod
    def inertia_tensor(self):
        pass

    def translate(self, inertia, displacement_vector):
        I_0 = inertia
        m = self.mass 
        R = displacement_vector
        return I_0 + m * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
    
    def rotate(self, inertia, rotation_matrix):
        I_0 = inertia
        R = rotation_matrix
        R_T = rotation_matrix.T
        return R @ I_0 @ R_T
    
    def rotate_to_vector(self, inertia, current_facing, to_face):
        
        if(np.all(np.isclose(current_facing, to_face))):
            return inertia
            
        rotation_axis = np.cross(to_face, current_facing)
        rotation_axis /= np.linalg.norm(rotation_axis)
        rotation_angle = np.arccos(np.dot(to_face, current_facing))

        K = np.array([
            [0, -rotation_axis[2], rotation_axis[1]],
            [rotation_axis[2], 0, -rotation_axis[0]],
            [-rotation_axis[1], rotation_axis[0], 0]
        ])
        R = np.eye(3) + K + np.dot(np.square(K), (1 - np.cos(rotation_angle))/np.sin(rotation_angle))

        return self.rotate(inertia, R)

    def shift_center(self, inertia, new_center):
        displacement = new_center - self.center
        return self.translate(inertia, displacement)



class Sphere(InertiaGeometry):

    def __init__(self, mass, center, radius):
        super().__init__(mass, center)
        self.radius = radius

    def inertia_tensor(self):
        I = (2 / 5) * self.mass * self.radius**2
        return I * np.eye(3)
    
class HollowCylinder(InertiaGeometry):

    def __init__(self, mass, center, inner_radius, outer_radius, height, facing):
        # facing is unit vector of flat side
        super().__init__(mass, center)
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.height = height
        self.facing = facing / np.linalg.norm(facing)

    def inertia_tensor(self):

        R = self.outer_radius
        k = self.inner_radius
        h = self.height

        I_zz = (1/2) * self.mass * (R**2 + k**2)  
        I_xx_yy = (1/4) * self.mass * (R**2 + k**2) + (1/12) * self.mass * h**2

        principal_tensor = np.diag([I_xx_yy, I_xx_yy, I_zz])

        return self.rotate_to_vector(principal_tensor, np.array([0, 0, 1]), self.facing)
    
class Cuboid(InertiaGeometry):

    def __init__(self, mass, center, width, height, depth):
        super().__init__(mass, center)
        self.width = width 
        self.height = height 
        self.depth = depth 

    def inertia_tensor(self):
        a = self.width
        b = self.height
        c = self.depth

        I_xx = (1 / 12) * self.mass * (b**2 + c**2)
        I_yy = (1 / 12) * self.mass * (a**2 + c**2)
        I_zz = (1 / 12) * self.mass * (a**2 + b**2)

        I = np.diag([I_xx, I_yy, I_zz])  # noqa: E741
        return I




class InertiaBuilder:
   
    def __init__(self, *args: InertiaGeometry):
        self.geometries = args

    def moment_of_inertia(self, center: np.ndarray = np.array([0, 0, 0])):

        total_inertia = np.zeros((3, 3))
        for geometry in self.geometries:
            shifted = geometry.shift_center(geometry.inertia_tensor(), center)
            total_inertia += shifted

        return total_inertia

# HollowCylinder(1, 1, 1, 1, 1, np.array([0, 0, 1])).inertia_tensor()


from ezauv.mission.mission import Subtask
from ezauv.hardware.sensor_interface import SensorInterface
from ezauv.simulation.core import Simulation

import numpy as np
import time

class Simulate(Subtask):

    def __init__(self, simulation: Simulation):
        super().__init__()
        self.simulation = simulation
        self.prevtime = -1.


    @property
    def name(self) -> str:
        return "Simulate subtask"

    def update(self, sensors: SensorInterface, wanted_speed: np.ndarray) -> np.ndarray:
        new_time = time.time()
        if(self.prevtime != -1):
            # set_text(str(new_time - self.prevtime))
            self.simulation.simulate(new_time - self.prevtime)
        self.prevtime = time.time()
        return np.array([0., 0., 0., 0., 0., 0.])
        


from abc import ABC, abstractmethod
from typing import Tuple
import numpy as np

class Task(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def finished(self) -> bool:
        pass

    @abstractmethod
    def update(self, sensors) -> np.ndarray:
        pass

class Subtask(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def update(self, sensors, wanted_speed: np.ndarray) -> np.ndarray:
        pass


class Path:
    def __init__(self, *args: Task):
        self.path: Tuple[Task, ...] = args



import time

# A PID controller is a piece of code which tries to bring a chaotic system to a specific
# value; eg. bring a sub to a specific depth or angle, maneuvering against the chaotic water.
# it is based on three parts; one Proportional, Integral, and Derivative (P-I-D). The final
# output signal is a weighted sum of the three parts. To find the proportional piece, you just
# take the error between the current value and the wanted value. The integral piece is the sum
# of all the previous errors. The derivative piece is more complicated, being the rate of change
# between the current and previous errors; so, (error - prev_error) / (time - prev_time).
# once you have these three pieces, the output is just a weighted sum of them, using
# Kp, Ki, and Kd respectively

# as of the 2023 competition, the parameters were
# Kp = 0.6
# Ki = 0.0
# Kd = 0.1
# these will probably (almost definitely) need to be retuned in the future


class PID:
    def __init__(self, proportional, integral, derivative, wanted, interval = None):
        """
        The first three parameters represent the weight of their respect part of the PID;
        Kp, Ki, and Kd. Wanted is the value the PID will attempt to reach, this cannot be
        changed without creating a new PID object. Interval is optional, if unset the PID
        will use real time, which is good in case of potential lag. If set, it will assume
        every call is `interval` seconds apart. It's good for testing, but in a real
        scenario use real time.
        """
        # the first three params represent the weight of their respective part of the PID
        # interval is optional, if unset the PID time will be based on actual time, which is 
        # good in case of potential lag. if set, it will assume every call is `interval`
        # seconds apart. it's good for testing, but in a real scenario use timestamp

        self.Kp = proportional
        self.Ki = integral
        self.Kd = derivative

        self.wanted = wanted

        self.interval = interval

        self.integral = 0
        # integral represents the total error

        self.prev = 0
        # prev represents the previous error

        self.prevTime = time.time()
        # only used if interval is unset

    def signal(self, current):
        """
        Get the control signal given the `current` value
        """
        # return the control signal, given a current value
        # this will update the saved data such as total error and previous error

        time_diff = self.interval if self.interval is not None else time.time() - self.prevTime
        error = self.wanted - current
        self.integral += error
        derivative = (error - self.prev) / time_diff

        self.prev = error
        signal = (self.Kp * error) + (self.Ki * self.integral) + (self.Kd * derivative)
        return signal


import numpy as np
from ezauv.hardware.sensor_interface import DepthInterface, ImuInterface


class FakeDepthSensor(DepthInterface):
        def __init__(self, deviation):
            self.deviation = deviation

        def get_depth(self):
            return 0.
        
        def initialize(self):
            pass

        def overview(self) -> str:
            return f"Simulated Depth Sensor -- Standard deviation: {self.deviation}. Currently hovercraft-only, no depth."
        
class FakeIMU(ImuInterface):

    def __init__(self, max_dev_accel, acceleration_function, rotation_function):
        self.max_dev_accel = max_dev_accel
        self.acceleration_function = acceleration_function
        self.rotation_function = rotation_function

    def get_accelerations(self):
        acceleration = self.acceleration_function() + (np.random.rand(2) - 0.5) * 2 * self.max_dev_accel
        return np.append(acceleration, 0) # bc it expects a 3d acceleration
    
    def get_rotation(self):
        return self.rotation_function()

    def initialize(self):
        pass

    def overview(self) -> str:
        return f"Simulated IMU -- Standard deviation: {self.max_dev_accel}"


import numpy as np
import traceback
from typing import Callable, List

from ezauv.hardware.motor_controller import MotorController
from ezauv.hardware.sensor_interface import SensorInterface
from ezauv.mission.mission import Path, Subtask
from ezauv.utils.logger import Logger, LogLevel
# from ezauv.simulation.animator import set_text

class AUV:
    def __init__(self, *,
                 motor_controller: MotorController, # the object to control the motors with
                 sensors: SensorInterface,          # the interface for sensor data
                 pin_kill: Callable,                # an emergency kill function; should disable all motors via pins

                 logging: bool = False,             # whether to save log to file
                 console: bool = True,              # whether to print log to console
                 ):
        """
        Create a sub wrapper object.\n
        motor_controller: the object to control the motors with\n
        sensors: the interface for all sensor data\n
        pin_kill: an emergency kill function, when the library is having issues. Should manually set motors off\n
        logging: whether to save log to file\n
        console: whether to print log to console
        """
        self.motor_controller = motor_controller
        self.sensors = sensors
        self.pin_kill = pin_kill

        self.logger = Logger(console, logging)

        self.motor_controller.log = self.logger.create_sourced_logger("MOTOR")
        self.sensors.log = self.logger.create_sourced_logger("SENSOR")
        self.sensors.depth.log = self.logger.create_sourced_logger("DEPTH")
        self.sensors.imu.log = self.logger.create_sourced_logger("IMU")

        self.logger.log("Sub enabled")
        self.motor_controller.overview()
        self.sensors.overview()

        self.motor_controller.initialize()
        self.sensors.initialize()

        self.subtasks: List[Subtask] = []

    def register_subtask(self, subtask):
        self.subtasks.append(subtask)

    def kill(self):
        self.motor_controller.set_motors(np.array([0 for _ in range(6)]))

    def travel_path(self, mission: Path) -> None:
        """Execute each Task in the given Path, in order, then kill the sub. Handles errors."""

        self.logger.log("Beginning path")

        try:
            for task in mission.path:
                self.logger.log(f"Beginning task {task.name}")
                # print(task.finished)
                while(not task.finished):
                    wanted_direction = np.copy(task.update(self.sensors))
                    
                    # total_subtask = np.array([0., 0., 0.])
                    # print(wanted_direction)
                    # print(np.sum([subtask.update(self.sensors, wanted_direction) for subtask in self.subtasks]))
                    # set_text(str(" ".join(str(t) for t in task.update(self.sensors))))
                    for subtask in self.subtasks:
                        wanted_direction += subtask.update(self.sensors, wanted_direction)
                    # wanted_direction += np.sum([subtask.update(self.sensors, wanted_direction) for subtask in self.subtasks])

                    solved_motors = self.motor_controller.solve(
                        wanted_direction,
                        self.sensors.imu.get_rotation()
                    )
                    # print(solved_motors[1])
                    if(solved_motors[0]):
                        # print("test")
                        self.motor_controller.set_motors(solved_motors[1])
                    

        except:
            self.logger.log(traceback.format_exc(), level=LogLevel.ERROR)
    
        finally:
            self.logger.log("Killing sub")


            if(not self.sensors.killed()):
                kill_methods = [
                ("kill", self.kill),
                # kill through sub interface, uses full library to send kill. should always work

                ("backup kill", self.pin_kill)
                # last resort, directly control pins and send kill commands. doesn't go through library
                # at all, just sends pin commands
                ]
                # when we get more kills (eg hardware kill once we connect it to raspi) add them here
            
                for method_name, method in kill_methods:
                    self.logger.log(f"Attempting {method_name}...")
                    method()
                    break
                    # if self.sensors.killed():
                        # self.logger.log(f"{method_name.capitalize()} succeeded")
                        # break
                    # else:
                        # self.logger.log(f"{method_name.capitalize()} failed", level=LogLevel.ERROR)
                else:
                    self.logger.log("All kills ineffective. Manual intervention required", level=LogLevel.ERROR)
            
            self.logger.end()


import math
import numpy as np
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"

import pygame
import imageio

debug_text = "test test"
def set_text(text):
    global debug_text
    debug_text = text

class SimulationAnimator:
    def __init__(self, width=600, height=600, fps=30, frames=100, output_dir="videos"):
        self.width = width
        self.height = height
        self.scale_factor = self.width / 30  
        self.origin = np.array([width // 2, height // 2])  
        self.fps = fps
        self.frames = frames
        self.output_dir = output_dir
        self.frame_count = 0

        os.makedirs(output_dir, exist_ok=True)
        self.screen = pygame.Surface((width, height))
        
        self.video_path = os.path.join(self.output_dir, "animation.mp4")
        self.writer = imageio.get_writer(self.video_path, fps=self.fps, codec='libx264', quality=8)
        
        pygame.font.init()
        self.font = pygame.font.SysFont('Arial', 30)

    def to_screen(self, unit_position):
        return (unit_position * self.scale_factor + self.origin).astype(int)

    def append(self, position, rotation, velocity, motor_positions, motor_accelerations):
        self.screen.fill((255, 255, 255)) 

        angle = math.radians(rotation)
        cos_a, sin_a = np.cos(angle), np.sin(angle)

        square_points = np.array([[-1, -1], [1, -1],
                                  [1, 1], [-1, 1]])
        rotated_points = [(cos_a * x - sin_a * y, sin_a * x + cos_a * y) for x, y in square_points]
        screen_points = [self.to_screen(np.array(p) + position) for p in rotated_points]

        pygame.draw.polygon(self.screen, (0, 0, 255), screen_points)  

        front_endpoint = position + np.array([cos_a, sin_a]) * 1.5
        pygame.draw.line(self.screen, (0, 0, 0), self.to_screen(position), self.to_screen(front_endpoint), 2)  

        velocity_endpoint = position + velocity * 0.5  
        pygame.draw.line(self.screen, (255, 0, 0), self.to_screen(position), self.to_screen(velocity_endpoint), 3)  

        for pos, acc in zip(motor_positions, motor_accelerations):
            pygame.draw.circle(self.screen, (200, 150, 0), self.to_screen(pos), 5)  
            acc_endpoint = pos + acc * 0.5  
            pygame.draw.line(self.screen, (200, 150, 0), self.to_screen(pos), self.to_screen(acc_endpoint), 2)  
        text_surface = self.font.render(debug_text, False, (0, 0, 0))
        self.screen.blit(text_surface, (0, 0))

        raw_frame = pygame.surfarray.array3d(self.screen)
        raw_frame = np.rot90(raw_frame, k=3)  # Rotate to match correct orientation
        raw_frame = np.fliplr(raw_frame)  # Flip horizontally
        self.writer.append_data(raw_frame)

    def render(self):
        self.writer.close()
        pygame.quit()
        print(f"Video saved at: {self.video_path}")



from ezauv.mission.mission import Subtask
from ezauv.hardware.sensor_interface import SensorInterface
from ezauv.utils.pid import PID

import numpy as np
import time
import quaternion



class HeadingPID(Subtask):

    def __init__(self, wanted_heading, Kp, Ki, Kd):
        super().__init__()
        self.pid = PID(Kp, Ki, Kd, 0)
        self.wanted = wanted_heading

    @property
    def name(self) -> str:
        return "Heading PID subtask"

    def update(self, sensors: SensorInterface, wanted_speed: np.ndarray) -> np.ndarray:
        q = sensors.imu.get_rotation()
        v_quat = q * np.quaternion(0, 0.6, 0.0, 0.1) * q.conjugate()
        # v
        v_prime = np.array([v_quat.x, v_quat.y, v_quat.z])

        # print(v_prime)
        yaw = np.degrees(np.arctan2(v_prime[1], v_prime[0]))

        diff = self.wanted - yaw
        # print(yaw)
        
        if abs(diff) >= 180:
            sign = yaw / abs(yaw)
            abs_diff_yaw = 180 - abs(yaw)
            abs_diff_target = 180 - abs(self.wanted)
            diff = sign * (abs_diff_yaw + abs_diff_target)

        signal = self.pid.signal(-diff)
        # print(quaternion.as_euler_angles(sensors.imu.get_rotation()))
        return np.array([0., 0., 0., 0., 0., signal])


from setuptools import setup, find_packages

setup(
    name='ezauv',
    version='0.1.3',
    packages=find_packages(),
    install_requires=[
        'gurobipy==12.0.1',
        'numpy==2.2.3',
        'pygame==2.6.1',
        'quaternion==3.5.2.post4',
        'scipy==1.15.2'
        ],
    description='A library to make coding AUVs easier',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Andre Gordon',
    author_email='gordona26@bcdschool.org',
    url='https://github.com/beaver-auv/ezauv',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License'
    ],
)
